/*
*         Portable Batch System (PBS) Software License
* 
* Copyright (c) 1999, MRJ Technology Solutions.
* All rights reserved.
* 
* Acknowledgment: The Portable Batch System Software was originally developed
* as a joint project between the Numerical Aerospace Simulation (NAS) Systems
* Division of NASA Ames Research Center and the National Energy Research
* Supercomputer Center (NERSC) of Lawrence Livermore National Laboratory.
* 
* Redistribution of the Portable Batch System Software and use in source
* and binary forms, with or without modification, are permitted provided
* that the following conditions are met:
* 
* - Redistributions of source code must retain the above copyright and
*   acknowledgment notices, this list of conditions and the following
*   disclaimer.
* 
* - Redistributions in binary form must reproduce the above copyright and 
*   acknowledgment notices, this list of conditions and the following
*   disclaimer in the documentation and/or other materials provided with the
*   distribution.
* 
* - All advertising materials mentioning features or use of this software must
*   display the following acknowledgment:
* 
*   This product includes software developed by NASA Ames Research Center,
*   Lawrence Livermore National Laboratory, and MRJ Technology Solutions.
* 
*         DISCLAIMER OF WARRANTY
* 
* THIS SOFTWARE IS PROVIDED BY MRJ TECHNOLOGY SOLUTIONS ("MRJ") "AS IS" WITHOUT 
* WARRANTY OF ANY KIND, AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
* BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
* FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT ARE EXPRESSLY DISCLAIMED.
* 
* IN NO EVENT, UNLESS REQUIRED BY APPLICABLE LAW, SHALL MRJ, NASA, NOR
* THE U.S. GOVERNMENT BE LIABLE FOR ANY DIRECT DAMAGES WHATSOEVER,
* NOR ANY INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
* OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
* CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
* OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
* USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
* 
* This license will be governed by the laws of the Commonwealth of Virginia,
* without reference to its choice of law rules.
*/
#include <pbs_config.h>   /* the master config generated by configure */

#include <sys/types.h>
#include <sys/param.h>
#include "pbs_ifl.h"
#include <stdio.h>
#include <grp.h>
#include <pwd.h>
#include <stdlib.h>
#include <string.h>
#include "server_limits.h"
#include "list_link.h"
#include "attribute.h"
#include "job.h"
#include "server.h"
#include "pbs_error.h"
#include "svrfunc.h"
#ifdef _CRAY
#include <udb.h>
#endif /* _CRAY */

static char ident[] = "@(#) $RCSfile: geteusernam.c,v $ $Revision: 2.1 $";

/* External Data */

extern char server_host[];

/*
 * geteusernam - get effective user name
 *	Get the user name under which the job should be run on this host:
 *	  1. from user_list use name with host name that matches this host
 *	  2. from user_list use name with no host specification
 *	  3. user owner name
 *
 *	Returns pointer to the user name
 */

static char *geteusernam(pjob, pattr)
	job       *pjob;
	attribute *pattr;	/* pointer to User_List attribute */
{
	int	 rule3 = 0;
	char	*hit = 0;
	int	 i;
	struct array_strings *parst;
	char	*pn;
	char	*ptr;
	static char username[PBS_MAXUSER+1];


	/* search the user-list attribute */

	if ( (pattr->at_flags & ATR_VFLAG_SET) &&
	     (parst = pattr->at_val.at_arst)  ) 	{
		for (i=0; i<parst->as_usedptr; i++) {
			pn = parst->as_string[i];
			ptr = strchr(pn, '@');
			if (ptr) {	/* has host specification */
			    if (!strncmp(server_host, ptr+1,strlen(ptr+1))) {
				hit = pn;	/* option 1. */
				break;
			    }
			} else {	/* wildcard host (null) */
			    hit = pn;	/* option 2. */
			}
		}
	}
	if (!hit) {	/* default to the job owner ( 3.) */
		hit = pjob->ji_wattr[(int)JOB_ATR_job_owner].at_val.at_str;
		rule3 = 1;
	}
	
	/* copy user name into return buffer, strip off host name */

	get_jobowner(hit, username);

	if (rule3) {
		(void)strcpy(username, site_map_user(username, get_variable(pjob, "PBS_O_HOST")));
	}
		
	return (username);
}

/*
 * getegroup - get specified effective group name
 *	Get the group name under which the job is specified to be run on
 *	this host:
 *	  1. from group_list use name with host name that matches this host
 *	  2. from group_list use name with no host specification
 *	  3. NULL, not specified.
 *
 *	Returns pointer to the group name or null
 */

static char *getegroup(pjob, pattr)
	job       *pjob;
	attribute *pattr;	/* group_list attribute */
{
	char	*hit = 0;
	int	 i;
	struct array_strings *parst;
	char	*pn;
	char	*ptr;
	static char groupname[PBS_MAXUSER+1];

	/* search the group-list attribute */

	if ( (pattr->at_flags & ATR_VFLAG_SET) &&
	     (parst = pattr->at_val.at_arst) )		{
		for (i=0; i<parst->as_usedptr; i++) {
			pn = parst->as_string[i];
			ptr = strchr(pn, '@');
			if (ptr) {	/* has host specification */
			    if (!strncmp(server_host, ptr+1,strlen(ptr+1))) {
				hit = pn;	/* option 1. */
				break;
			    }
			} else {	/* wildcard host (null) */
			    hit = pn;	/* option 2. */
			}
		}
	}
	if (!hit) 	/* nothing sepecified, return null */
		return ((char *)0);

	/* copy group name into return buffer, strip off host name */
	/* get_jobowner() works for group as well, same size	   */

	get_jobowner(hit, groupname);
	return (groupname);
}

/*
 * set_jobexid - validate and set the execution uid (JOB_ATR_euser) and
 *	execution group (JOB_ATR_egroup) job attributes.
 *	
 *	1.  Get the execution user name.
 *	1a. Get the corresponding password entry.
 *	1b. Uid of 0 (superuser) is not allowed, might cause root-rot
 *	1c. If job owner name is not being used, does the job owner have
 *	    permission to map to the execution name?
 *	1d. Set JOB_ATR_euser to the execution user name.
 *	2.  Get the execution group name.
 *	2a. Is the execution user name a member of this group?
 *	2b. Set JOB_ATR_egroup to the execution group name.
 *
 * Returns: 0 if set, non-zero error number if any error
 */

int set_jobexid(pjob, attrry)
	job       *pjob;
	attribute *attrry;
{
	int		 addflags = 0;
	attribute	*pattr;
	char	       **pmem;
	struct group	*gpent;
	char		*puser;
	struct passwd	*pwent;
	char		*pgrpn;
	char		 gname[PBS_MAXGRPN+1];
#ifdef _CRAY
	struct udb *pudb;
#endif

	/* use the passed User_List if set, may be a newly modified one     */
	/* if not set, fall back to the job's actual User_List, may be same */

	if ((attrry + (int)JOB_ATR_userlst)->at_flags & ATR_VFLAG_SET)
		pattr = attrry + (int)JOB_ATR_userlst;
	else
		pattr = &pjob->ji_wattr[(int)JOB_ATR_userlst];
	if ((puser = geteusernam(pjob, pattr)) == (char *)0)
		return (PBSE_BADUSER);

	pwent = getpwnam(puser);
	if (pwent == (struct passwd *)0) 
		return (PBSE_BADUSER);
	if (pwent->pw_uid == 0) {
	    if (server.sv_attr[(int)SRV_ATR_AclRoot].at_flags & ATR_VFLAG_SET) {
		if (acl_check(&server.sv_attr[(int)SRV_ATR_AclRoot],
			   pjob->ji_wattr[(int)JOB_ATR_job_owner].at_val.at_str,
			   ACL_User) == 0)
				return (PBSE_BADUSER); /* root not allowed */
	    } else {
		return (PBSE_BADUSER); /* root not allowed */
	    }
	}

	if (site_check_user_map(pjob, puser) == -1)
		return (PBSE_BADUSER);

	pattr = attrry + (int)JOB_ATR_euser;
	job_attr_def[(int)JOB_ATR_euser].at_free(pattr);
	job_attr_def[(int)JOB_ATR_euser].at_decode(pattr, (char *)0, (char *)0,
						   puser);
	
#ifdef _CRAY
	/* On cray check UDB (user data base) for permission to batch it    */

	pudb = getudbuid(pwent->pw_uid);
	endudb();
	if (pudb == UDB_NULL) 
		return (PBSE_BADUSER);
	if (pudb->ue_permbits & (PERMBITS_NOBATCH | PERMBITS_RESTRICTED))
		return (PBSE_QACESS);

	/* if account (qsub -A) not specified, set default from UDB */

	pattr = attrry + (int)JOB_ATR_account;
	if ((pattr->at_flags & ATR_VFLAG_SET) == 0) {
		(void)job_attr_def[(int)JOB_ATR_account].at_decode( pattr,
			(char *)0, (char *)0,
			(char *)acid2nam(pudb->ue_acids[0]) );
	}
#endif	/* _CRAY */

	/*
	 * now figure out the group name under which the job should execute
	 * PBS requires that each group have an entry in the group file,
	 * see the admin guide for the reason why...
	 *
	 * use the passed group_list if set, may be a newly modified one
	 * if not set, fall back to the job's actual group_list, may be same
	 */

	if ((attrry + (int)JOB_ATR_grouplst)->at_flags & ATR_VFLAG_SET)
		pattr = attrry + (int)JOB_ATR_grouplst;
	else
		pattr = &pjob->ji_wattr[(int)JOB_ATR_grouplst];
	if (pgrpn = getegroup(pjob, pattr)) {	

		/* user specified a group, group must exists and either	   */
		/* must be user's primary group	 or the user must be in it */

		gpent = getgrnam(pgrpn);
		if (gpent == (struct group *)0)
			return (PBSE_BADGRP);		/* no such group */

		if (gpent->gr_gid != pwent->pw_gid) {	/* not primary group */
			pmem = gpent->gr_mem;
			while (*pmem) {
				if ( !strcmp(puser, *pmem) )
					break;
				++pmem;
			}
			if (*pmem == 0)
				return (PBSE_BADGRP);	/* user not in group */
		}

	} else {

		/* Use user login group */

		gpent = getgrgid(pwent->pw_gid);
		if (gpent != (struct group *)0) {
			pgrpn = gpent->gr_name;		/* use group name */
		} else {
			(void)sprintf(gname, "%d", pwent->pw_gid);
			pgrpn = gname;		/* turn gid into string */
		}

		/*
		 * setting the DEFAULT flag is a "kludy" way to keep MOM from
		 * having to do an unneeded look up of the group file.
		 * We needed to have JOB_ATR_egroup set for the server but
		 * MOM only wants it if it is not the login group, so there!
		 */
		addflags = ATR_VFLAG_DEFLT;
	}


	pattr = attrry + (int)JOB_ATR_egroup;
	job_attr_def[(int)JOB_ATR_egroup].at_free(pattr);

	job_attr_def[(int)JOB_ATR_egroup].at_decode(pattr, (char *)0,
						(char *)0, pgrpn);
	pattr->at_flags |= addflags;
	return (0);
}
